/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.List;
import java.util.logging.Level;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.stream.StreamSource;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.commons.stream.InputStreamForker;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.AbstractAsyncContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;
import org.ow2.petals.se.eip.async.DynamicRouterAsyncContext;
import org.w3c.dom.Document;

/**
 * A Router pattern
 * 
 * @author alouis
 * @author ofabre
 * @author Christophe HAMERLING - eBM WebSourcing
 * @author Frederic Gardes
 * @since 2.2
 */
public class DynamicRouter extends AbstractMEPBridgePattern {

    public final static String DYNAMIC_ROUTER_NAME = "Dynamic-Router";

    public final static String DYNAMIC_ROUTER_NAMESPACE = EIPConstants.NAMESPACE + "/dynamicRouter";

    private RouterEngine routerEngine = null;

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.util.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        try {
            final List<Consumes> consumesList = context.getSUConsumes(exchange.getEndpoint());
            if (consumesList != null) {
                final List<String> conditions = this.routerEngine.retrieveConditions(context
                        .getExtensions());
                // there must be numberOfConditions+2 consumes defined in the SU
                if (consumesList.size() == conditions.size() + 2) {
                    // the first consumes is the targeted service used to
                    // evaluate the
                    // condition
                    final Consumes conditionService = consumesList.get(0);
                    if (conditionService != null) {
                        // Handle the Streaming issue
                        final InputStreamForker inMessageAsStream = PatternHelper
                        .handleStreamSource(exchange.getInMessage());
                        final Exchange conditionExchange = context
                        .createConsumeExchange(conditionService);
                        PatternHelper.copy(exchange.getInMessage(), conditionExchange
                                .getInMessage());
                        // Reset the Source to the in if necessary
                        if (inMessageAsStream != null) {
                            exchange.getInMessage().setContent(
                                    new StreamSource(inMessageAsStream.getInputStreamTwo()));
                        }
                        UtilFactory.getExchangeUtil().copyExchangeProperties(exchange,
                                conditionExchange);
                        final boolean faultRobust = this
                        .isFaultToException(context.getExtensions());
                        final DynamicRouterAsyncContext asyncContext = new DynamicRouterAsyncContext(
                                exchange, AbstractAsyncContext.DEFAULT_TIME_TO_LIVE, faultRobust,
                                consumesList, conditions);
                        context.sendAsync(conditionExchange, asyncContext);
                    } else {
                        exchange.setError(new JBIException(this.getName()
                                + ": the Service Unit must define a valid condition service"));
                    }
                } else {
                    exchange
                    .setError(new JBIException(
                            this.getName()
                            + ": the Service Unit must define (numberOfConditions+2) Consumes sections for the pattern"));
                }
            } else {
                exchange
                .setError(new JBIException(
                        this.getName()
                        + ": the Service Unit must define (numberOfConditions+2) Consumes sections for the pattern"));
            }
        } catch (final JBIException e) {
            exchange.setError(new Exception(e));
        }
        return false;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return DYNAMIC_ROUTER_NAME;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    public boolean validateMEP(final URI mep) {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
        this.routerEngine = new RouterEngine();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#processAsync(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext,
     * org.ow2.petals.se.eip.async.CommonAsyncContext)
     */
    @Override
    public boolean processAsync(final Exchange exchange, final ExchangeContext context,
            final CommonAsyncContext asyncContext) {
        try {
            final int subMessages = asyncContext.decrementAndGetOngoingSubMessages();
            if (subMessages == 1) {
                this.processRouterResponse(exchange, context, asyncContext);
            } else if (subMessages == 0) {
                this.processResponse(exchange, context, asyncContext);
            } else {
                if (!asyncContext.isTerminated()) {
                    asyncContext
                    .getOriginalExchange()
                    .setError(
                            new Exception(
                            "Terminate the pattern processing due to a bad global routing"));
                    context.send(asyncContext.getOriginalExchange());
                    asyncContext.setTerminated();
                }
            }
        } catch (final MessagingException e) {
            this.logger.warning(this.getName()
                    + ": unable to close the source exchange on error status");
        }
        return false;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getNameSpace()
     */
    @Override
    public String getNameSpace() {
        return DYNAMIC_ROUTER_NAMESPACE;
    }

    private void processRouterResponse(final Exchange exchange, final ExchangeContext context,
            final CommonAsyncContext asyncContext) {
        NormalizedMessage receivedMessage = null;
        Document document = null;
        try {
            // process the response
            receivedMessage = PatternHelper.processSentExchange(exchange, context);
            if (asyncContext instanceof DynamicRouterAsyncContext) {
                if (!asyncContext.isTerminated()) {
                    // For the source exchange, use the original MEP
                    if (this.logger != null && this.logger.isLoggable(Level.FINE)) {
                        this.logger.fine(this.getName() + ": processing the dynamic pattern");
                    }
                    if (receivedMessage != null) {
                        // the fault or the out message is processed
                        document = UtilFactory.getSourceUtil().createDocument(
                                receivedMessage.getContent());
                        // retrieve identifier of the first consumes linked to a
                        // true
                        // condition, or the default one
                        final int targetedConsumesListIndentifier = this.routerEngine
                        .retrieveTargetedConsumesListIdentifier(
                                ((DynamicRouterAsyncContext) asyncContext)
                                .getConditionList(), document, context);
                        // forward the exchange to the correct service,
                        // depending on
                        // the
                        // test result
                        // skip the router consumes
                        final Consumes consumedService = ((DynamicRouterAsyncContext) asyncContext)
                        .getConsumesList().get(targetedConsumesListIndentifier + 1);
                        final Exchange routedExchange = context
                        .createConsumeExchange(consumedService);
                        PatternHelper.copy(asyncContext.getOriginalExchange().getInMessage(),
                                routedExchange.getInMessage());
                        UtilFactory.getExchangeUtil().copyExchangeProperties(
                                asyncContext.getOriginalExchange(), routedExchange);
                        this.logSend(consumedService, routedExchange);
                        context.sendAsync(routedExchange, asyncContext);
                    } else {
                        asyncContext
                        .getOriginalExchange()
                        .setError(
                                new Exception(
                                "No valid response from the dynamic router service: unable to route"));
                        // in this case, it's an error, terminate the original
                        // exchange
                        try {
                            context.send(asyncContext.getOriginalExchange());
                        } catch (final MessagingException e1) {
                            this.logger.warning(this.getName()
                                    + ": unable to close the source exchange on error status");
                        } finally {
                            asyncContext.setTerminated();
                        }
                    }
                }
            } else if (!asyncContext.isTerminated()) {
                asyncContext
                .getOriginalExchange()
                .setError(
                        new Exception(
                        "The asynchronous context isn't a dynamic router context: unable to process"));
                // in this case, it's an error, terminate the original exchange
                try {
                    context.send(asyncContext.getOriginalExchange());
                } catch (final MessagingException e1) {
                    this.logger.warning(this.getName()
                            + ": unable to close the source exchange on error status");
                } finally {
                    asyncContext.setTerminated();
                }
            }
        } catch (final JBIException e) {
            if (!asyncContext.isTerminated()) {
                asyncContext.getOriginalExchange().setError(new Exception(e));
                // in this case, it's an error, terminate the original exchange
                try {
                    context.send(asyncContext.getOriginalExchange());
                } catch (final MessagingException e1) {
                    this.logger.warning(this.getName()
                            + ": unable to close the source exchange on error status");
                } finally {
                    asyncContext.setTerminated();
                }
            }
        } catch (final XPathExpressionException e) {
            if (!asyncContext.isTerminated()) {
                asyncContext.getOriginalExchange().setError(new Exception(e));
                // in this case, it's an error, terminate the original exchange
                try {
                    context.send(asyncContext.getOriginalExchange());
                } catch (final MessagingException e1) {
                    this.logger.warning(this.getName()
                            + ": unable to close the source exchange on error status");
                } finally {
                    asyncContext.setTerminated();
                }
            }
        }
    }
}
