/**
 * The Splitter EIP pattern
 */

package org.ow2.petals.se.eip.patterns;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;

import javax.activation.DataHandler;
import javax.activation.DataSource;
import javax.jbi.JBIException;
import javax.jbi.messaging.MessagingException;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.StringHelper;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.AbstractAsyncContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;

/**
 * Split the source exchange and send the parts to a target service.
 * 
 * @author Adrien LOUIS - EBMWebSourcing
 * @author Frederic GARDES
 * 
 */
/**
 * @author fgardes
 */
public class Splitter extends AbstractForkerPattern {

    private final static String SPLITTER_NAMESPACE = EIPConstants.NAMESPACE + "/splitter";

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return "Splitter";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        boolean send = false;
        boolean reply = false;
        try {
            final List<Consumes> consumesList = context.getSUConsumes(exchange.getEndpoint());
            if (consumesList != null && consumesList.size() == 1) {
                final Consumes consumes = consumesList.get(0);
                List<Source> splittedSources = null;
                if (!this.isAttachmentMode(context.getExtensions())) {
                    // content mode
                    final String splitterPath = context.getExtensions().get(EIPConstants.TEST);
                    if (!StringHelper.isNullOrEmpty(splitterPath)) {
                        if (this.logger != null && this.logger.isLoggable(Level.FINE)) {
                            this.logger
                            .fine(this.getName() + ": splitting XPath : " + splitterPath);
                        }
                        final Document inMessageContent = UtilFactory.getSourceUtil()
                        .createDocument(exchange.getInMessage().getContent());
                        final NodeList nodeList = PatternHelper.evaluateXPathExpressionAsNodeList(
                                splitterPath, inMessageContent);
                        if (nodeList != null && nodeList.getLength() != 0) {
                            if (this.logger != null && this.logger.isLoggable(Level.FINE)) {
                                this.logger.fine(this.getName() + ": splitted elements : "
                                        + nodeList.getLength());
                            }
                            splittedSources = new ArrayList<Source>(nodeList.getLength());
                            for (int i = 0; i < nodeList.getLength(); i++) {
                                splittedSources.add(new DOMSource(nodeList.item(i)));
                                send = true;
                            }
                        } else {
                            exchange.setError(new JBIException(this.getName()
                                    + ": no element to process found"));
                            reply = true;
                        }
                    } else {
                        exchange
                        .setError(new JBIException(this.getName()
                                + ": the Service Unit must define an element "
                                + EIPConstants.TEST));
                        reply = true;
                    }
                } else {
                    // attachment mode
                    if (exchange.getInMessageAttachments() != null
                            && exchange.getInMessageAttachments().size() != 0) {
                        splittedSources = new ArrayList<Source>(exchange.getInMessageAttachments()
                                .size());
                        for (final DataHandler dataHandler : exchange.getInMessageAttachments()) {
                            final String name = dataHandler.getName();
                            final String type = dataHandler.getContentType();
                            if (this.logger != null && this.logger.isLoggable(Level.FINE)) {
                                this.logger.fine(this.getName()
                                        + ": type of the current attachment " + name + ": " + type);
                            }
                            if ("application/xml".equals(type)) {
                                final DataSource content = dataHandler.getDataSource();
                                splittedSources.add(new StreamSource(content.getInputStream()));
                                send = true;
                            } else if (this.logger != null) {
                                this.logger
                                .info(this.getName()
                                        + ": the attachment "
                                        + name
                                        + " with "
                                        + type
                                        + " as type is ignored. It must be an application/xml type to be process ");
                            }
                        }
                    } else if (this.logger != null) {
                        this.logger.info(this.getName() + ": no attachment to process found");
                    }
                }
                // check if we have to send
                if (send) {
                    final boolean faultRobust = this.isFaultRobust(context.getExtensions());
                    final boolean errorRobust = this.isExceptionRobust(context.getExtensions());
                    final CommonAsyncContext asyncContext = new CommonAsyncContext(exchange,
                            AbstractAsyncContext.DEFAULT_TIME_TO_LIVE, this.getName(),
                            splittedSources.size(), faultRobust, errorRobust, false);
                    for (final Source splittedSource : splittedSources) {
                        final Exchange splittedExchange = context.createConsumeExchange(consumes);
                        PatternHelper
                        .copy(exchange.getInMessage(), splittedExchange.getInMessage());
                        splittedExchange.getInMessage().setContent(splittedSource);
                        UtilFactory.getExchangeUtil().copyExchangeProperties(exchange,
                                splittedExchange);
                        this.logSend(consumes, splittedExchange);
                        context.sendAsync(splittedExchange, asyncContext);
                    }
                }
            } else {
                exchange.setError(new JBIException(this.getName()
                        + ": the Service Unit must define one and only one consumes section"));
                reply = true;
            }
        } catch (final MessagingException e) {
            exchange.setError(new Exception(e));
            reply = true;
        } catch (final XPathExpressionException e) {
            exchange.setError(new Exception(e));
            reply = true;
        } catch (final IOException e) {
            exchange.setError(new Exception(e));
            reply = true;
        } catch (final PEtALSCDKException e) {
            exchange.setError(new Exception(e));
            reply = true;
        }
        return reply;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractForkerPattern#getNameSpace()
     */
    @Override
    public String getNameSpace() {
        return SPLITTER_NAMESPACE;
    }
}
