/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.List;
import java.util.logging.Level;

import javax.jbi.JBIException;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.commons.stream.InputStreamForker;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.StringHelper;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;

/**
 * Before sending the Exchange to the Service provider (or consumer for the response), Copy the In or OUT message and send it to a "monitoring service". <br>
 * The message sent to the monitoring service is InOnly <br>
 * <li>request mode: copy the content of the in message to the monitor and send the exchange to the provider; response is received and send back to the consumer</li> <li>response mode : send the exchange to the provider, receive the response, copy the content of the out or the fault message to the monitor and send back
 * the response to the consumer</li> <li>request-response mode : send the exchange to the provider, receive the response, copy the content of the in and the out or the fault messages to the monitor</li> <li>request-on-response mode : send the exchange to the provider, receive the response, copy the content of the in
 * message if no fault message received to the monitor</li><br />
 * TODO This pattern must be refactored to be asynchronous
 * 
 * @author alouis, chamerling - eBM WebSourcing
 * @author Frederic Gardes
 */
public class WireTap extends AbstractPattern {

	public final static String WIRE_TAP_NAMESPACE = EIPConstants.NAMESPACE + "/wireTap";

	public static final String REQUEST_WAY = "request";

	public static final String RESPONSE_WAY = "response";

	public static final String REQUEST_RESPONSE_WAY = "request-response";

	public static final String REQUEST_ON_RESPONSE_WAY = "request-on-response";

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.Pattern#processPattern(org.ow2.petals. component.framework.util.Exchange, org.ow2.petals.se.eip.ExchangeContext)
	 */
	@Override
	public boolean process(final Exchange exchange, final ExchangeContext context) {
		final List<Consumes> consumesList = context.getSUConsumes(exchange.getEndpoint());
		if (consumesList.size() == 2) {
			if (!StringHelper.isNullOrEmpty(context.getExtensions().get(EIPConstants.WIRE_TAP_WAY))) {
				this.sendExchange(consumesList, exchange, context);
			} else {
				exchange.setError(new JBIException(this.getName() + ": parameter '" + EIPConstants.WIRE_TAP_WAY + "' is mandatory for the pattern "));
			}
		} else {
			exchange.setError(new JBIException(this.getName() + ": the Service Unit must define 2 Consumes sections for the pattern " + this.getName()));
		}
		return true;
	}

	/**
	 * Send the exchange to its destination and send a copy to the 'monitoring service'. <br>
	 * Depending on the "way" parameter , copy the IN or OUT( or fault) message and send it to the 'monitoring service', as InOnly exchange. <br>
	 * values for "way" parameter : <br>
	 * "request" send the IN message (InOnly MEP) (default mode) <br>
	 * "response" send the OUT or Fault or ERROR as an IN message (InOnly MEP) default mode for "way" is "request"
	 * 
	 * @param consumesList
	 * @param exchange
	 * @param context
	 */
	private void sendExchange(final List<Consumes> consumesList, final Exchange exchange, final ExchangeContext context) {
		Exchange monitorInExchange = null;
		Exchange monitorOutExchange = null;

		try {
			final Consumes provider = consumesList.get(0);
			final Consumes monitor = consumesList.get(1);
			final String way = context.getExtensions().get(EIPConstants.WIRE_TAP_WAY);
			if (REQUEST_WAY.equals(way) || REQUEST_RESPONSE_WAY.equals(way) || REQUEST_ON_RESPONSE_WAY.equals(way)) {
				monitorInExchange = this.prepareRequestCopyIfNeeded(monitor, exchange, context);
			}
			// send IN to the monitor if Request way is defined
			if (monitorInExchange != null && (REQUEST_WAY.equals(way) || REQUEST_RESPONSE_WAY.equals(way))) {
				if (this.logger != null && this.logger.isLoggable(Level.FINE)) {
					this.logger.fine(this.getName() + ": invocation of the in monitoring service : interface name=" + monitor.getInterfaceName().toString() + " / service name=" + monitor.getServiceName().toString() + " / endpoint name=" + monitor.getEndpointName());
				}
				context.send(monitorInExchange);
			}
			// send (sync) exchange to provider
			final Exchange targetExchange = this.sendSyncExchange(provider, exchange, context);
			// send IN to the monitor if RequestOnResponse way is defined
			// and response is OK (no fault or error)
			if (monitorInExchange != null && REQUEST_ON_RESPONSE_WAY.equals(way) && targetExchange.getError() == null && targetExchange.getFault() == null) {
				if (this.logger != null && this.logger.isLoggable(Level.FINE)) {
					this.logger.fine(this.getName() + ": invocation of the in monitoring service : interface name=" + monitor.getInterfaceName().toString() + " / service name=" + monitor.getServiceName().toString() + " / endpoint name=" + monitor.getEndpointName());
				}
				context.send(monitorInExchange);
			}
			if (RESPONSE_WAY.equals(way) || REQUEST_RESPONSE_WAY.equals(way)) {
				monitorOutExchange = this.prepareResponseCopyIfNeeded(monitor, targetExchange, context);
				// send OUT, Fault or error to the monitor if Response way is
				// defined
				if (monitorOutExchange != null) {
					if (this.logger != null && this.logger.isLoggable(Level.FINE)) {
						this.logger.fine(this.getName() + ": invocation of the out monitoring service : interface name=" + monitor.getInterfaceName().toString() + " / service name=" + monitor.getServiceName().toString() + " / endpoint name=" + monitor.getEndpointName());
					}
					context.send(monitorOutExchange);
				}
			}
			this.prepareSourceExchange(PatternHelper.processSentExchange(targetExchange, context), exchange);
		} catch (final JBIException e) {
			if (exchange.isActiveStatus()) {
				exchange.setError(new Exception(e));
			}
		}
	}

	/**
	 * Copy into the source exchange the content of the response of the target exchange
	 * 
	 * @param receivedMessage
	 * @param exchange
	 * @throws MessagingException
	 * @throws PEtALSCDKException
	 */
	private void prepareSourceExchange(final NormalizedMessage receivedMessage, final Exchange exchange) throws MessagingException, PEtALSCDKException {
		if (exchange.isActiveStatus()) {
			// process the response
			if (receivedMessage != null) {
				if (receivedMessage instanceof Fault) {
					if (!exchange.isInOnlyPattern()) {
						final Fault fault = exchange.createFault();
						PatternHelper.copy(receivedMessage, fault);
						exchange.setFault(fault);
					}
				} else {
					if (!exchange.isInOnlyPattern() && !exchange.isRobustInOnlyPattern()) {
						PatternHelper.copy(receivedMessage, exchange.getOutMessage());
					}
				}
			} else if (exchange.isInOutPattern()) {
				exchange.setError(new Exception(this.getName() + ": the target exchange doesn't contain an out message: unable to process"));
			}
		}
	}

	/**
	 * @param monitor
	 * @param exchange
	 * @param context
	 * @return the exchange for the monitor, null if not needed
	 * @throws PEtALSCDKException
	 * @throws JBIException
	 */
	private Exchange prepareRequestCopyIfNeeded(final Consumes monitor, final Exchange exchange, final ExchangeContext context) throws PEtALSCDKException, JBIException {
		Exchange monitorExchange = null;

		if (exchange.getInMessage() != null) {
			// Handle the Streaming issue if necessary
			final InputStreamForker inMessageAsStream = PatternHelper.handleStreamSource(exchange.getInMessage());
			if (inMessageAsStream == null) {
				final Source inContent = exchange.getInMessage().getContent();
				if (inContent == null) {
					throw new PEtALSCDKException("The content must be a StreamSource or a DOMSource");
				}
				if (!(inContent instanceof DOMSource)) {
					throw new PEtALSCDKException("The content must be a StreamSource or a DOMSource");
				}
			}
			monitorExchange = context.createConsumeExchange(monitor);
			PatternHelper.copy(exchange.getInMessage(), monitorExchange.getInMessage());
			UtilFactory.getExchangeUtil().copyExchangeProperties(exchange, monitorExchange);
			// Reset the Source to the InMessage if necessary
			if (inMessageAsStream != null) {
				exchange.getInMessage().setContent(new StreamSource(inMessageAsStream.getInputStreamTwo()));
			}
		}
		return monitorExchange;
	}

	/**
	 * Send the exchange to the provider (only set the information, the framework will send it later)
	 * 
	 * @param provider
	 * @param exchange
	 * @param context
	 * @return the target exchange
	 * @throws JBIException
	 */
	private Exchange sendSyncExchange(final Consumes provider, final Exchange exchange, final ExchangeContext context) throws JBIException {
		Exchange providerExchange = null;
		providerExchange = context.createConsumeExchange(provider);
		PatternHelper.copy(exchange.getInMessage(), providerExchange.getInMessage());
		UtilFactory.getExchangeUtil().copyExchangeProperties(exchange, providerExchange);
		this.logSend(provider, providerExchange);
		context.sendSync(providerExchange);
		return providerExchange;
	}

	/**
	 * @param monitor
	 * @param targetExchange
	 *            The target exchange
	 * @param context
	 * @return The exchange for the monitor, null if not needed
	 * @throws PEtALSCDKException
	 * @throws JBIException
	 */
	private Exchange prepareResponseCopyIfNeeded(final Consumes monitor, final Exchange targetExchange, final ExchangeContext context) throws PEtALSCDKException, JBIException {
		Exchange monitorExchange = null;
		if (targetExchange.isActiveStatus()) {
			if (targetExchange.getFault() != null) {
				// Handle the Streaming issue if necessary
				final InputStreamForker faultMessageAsStream = PatternHelper.handleStreamSource(targetExchange.getFault());
				if (faultMessageAsStream == null) {
					final Source faultContent = targetExchange.getFault().getContent();
					if (faultContent == null) {
						throw new PEtALSCDKException("The fault must be a StreamSource or a DOMSource");
					}
					if (!(faultContent instanceof DOMSource)) {
						throw new PEtALSCDKException("The fault must be a StreamSource or a DOMSource");
					}
				}
				monitorExchange = context.createConsumeExchange(monitor);
				PatternHelper.copy(targetExchange.getFault(), monitorExchange.getInMessage());
				UtilFactory.getExchangeUtil().copyExchangeProperties(targetExchange, monitorExchange);
				// Reset the Source to the fault if necessary
				if (faultMessageAsStream != null) {
					targetExchange.getFault().setContent(new StreamSource(faultMessageAsStream.getInputStreamTwo()));
				}
			} else if (targetExchange.getOutMessage() != null) {
				// Handle the Streaming issue if necessary
				final InputStreamForker outMessageAsStream = PatternHelper.handleStreamSource(targetExchange.getOutMessage());
				if (outMessageAsStream == null) {
					final Source outContent = targetExchange.getOutMessage().getContent();
					if (outContent == null) {
						throw new PEtALSCDKException("The out must be a StreamSource or a DOMSource");
					}
					if (!(outContent instanceof DOMSource)) {
						throw new PEtALSCDKException("The out must be a StreamSource or a DOMSource");
					}
				}
				monitorExchange = context.createConsumeExchange(monitor);
				PatternHelper.copy(targetExchange.getOutMessage(), monitorExchange.getInMessage());
				UtilFactory.getExchangeUtil().copyExchangeProperties(targetExchange, monitorExchange);
				// Reset the Source to the out if necessary
				if (outMessageAsStream != null) {
					targetExchange.getOutMessage().setContent(new StreamSource(outMessageAsStream.getInputStreamTwo()));
				}
			}
		} else if (targetExchange.isErrorStatus()) {
			if (targetExchange.getError() != null) {
				monitorExchange = context.createConsumeExchange(monitor);
				monitorExchange.setError(targetExchange.getError());
				UtilFactory.getExchangeUtil().copyExchangeProperties(targetExchange, monitorExchange);
			}
		}
		return monitorExchange;
	}

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.Pattern#init()
	 */
	public void init() {
	}

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
	 */
	public String getName() {
		return "Wire-Tap";
	}

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
	 */
	@Override
	public boolean validateMEP(final URI mep) {
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.AbstractPattern#processAsync(org.ow2.petals .component.framework.api.message.Exchange, org.ow2.petals.se.eip.ExchangeContext, org.ow2.petals.se.eip.async.CommonAsyncContext)
	 */
	@Override
	public boolean processAsync(final Exchange exchange, final ExchangeContext context, final CommonAsyncContext asyncContext) {
		exchange.setError(new Exception(this.getName() + ": the pattern doesn't process the asynchronous exchanges"));
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getNameSpace()
	 */
	@Override
	public String getNameSpace() {
		return WIRE_TAP_NAMESPACE;
	}
}
